/*
 *  dhe/test.c
 *  tests the DHE computations
 *  samuel judson | cryptography and network protocols | marlboro college
 */

#include <stdio.h>
#include <string.h>
#include <gmp.h>

#include "dhe.h"

#ifndef DHE_SPECS
#include "../dhe/dhe_specs.h"
#define DHE_SPECS 1
#endif

#define HEX_CHAR_PER_BYTE 2


/*
 *  DHE Test
 */


void dhe_test() {
  char sec_str[DHE_KEY_LENGTH * HEX_CHAR_PER_BYTE];

  mpz_t secret_key;
  mpz_t generator;
  mpz_t prime;
  mpz_t other_party_secret;

  mpz_init_set_ui(secret_key, (unsigned long)6);
  mpz_init_set_ui(generator, (unsigned long)5);
  mpz_init_set_ui(prime, (unsigned long)23);
  mpz_init_set_ui(other_party_secret, (unsigned long)19);

  mpz_t party_secret;
  mpz_t shared_secret;

  mpz_t party_secret_expected;
  mpz_t shared_secret_expected;

  mpz_init(party_secret);
  mpz_init(shared_secret);

  mpz_init_set_ui(party_secret_expected, (unsigned long)8);
  mpz_init_set_ui(shared_secret_expected, (unsigned long)2);

  dhe_compute_party_secret(party_secret, secret_key, generator, prime);
  dhe_compute_shared_secret(shared_secret, secret_key, other_party_secret, prime);
  
  int passes = 1;

  fprintf(stdout, "\n===== DHE TEST =====\n\n");

  gmp_fprintf(stdout, "=== Components ===\n\n");
  gmp_fprintf(stdout, "Secret Key: %Zd\n", secret_key);
  gmp_fprintf(stdout, "Generator: %Zd\n", generator);
  gmp_fprintf(stdout, "Prime: %Zd\n", prime);
  gmp_fprintf(stdout, "Other Party Secret: %Zd\n\n---\n\n", other_party_secret);

  gmp_fprintf(stdout, "Party Secret: %Zd\n", party_secret);
  if (!mpz_cmp(party_secret, party_secret_expected)) {
    fprintf(stdout, "Party Secret is as expected.\n\n");
  } else {
    fprintf(stdout, "Party Secret is not as expected.\n\n");
    passes = 0;
  }

  gmp_fprintf(stdout, "Shared Secret: %Zd\n", shared_secret);
  if (!mpz_cmp(shared_secret, shared_secret_expected)) {
    fprintf(stdout, "Shared Secret is as expected.\n\n");
  } else {
    fprintf(stdout, "Shared Secret is not as expected.\n\n");
    passes = 0;
  }

  if (passes) {
    fprintf(stdout, "=== DHE TEST: PASS ===\n\n");
  } else {
    fprintf(stdout, "=== DHE TEST: FAIL ===\n\n");
  }
}


int main() {
  dhe_test();
};
