/*
 *  dhe/test.c
 *  tests the DHE computations
 *  samuel judson | cryptography and network protocols | marlboro college
 */

#include <stdio.h>
#include <string.h>
#include <gmp.h>

#include "rsa.h"

#ifndef RSA_SPECS
#include "../rsa/rsa_specs.h"
#define RSA_SPECS 1
#endif

#define HEX_CHAR_PER_BYTE 2


/*
 *  RSA Test
 */


void rsa_test() {
  mpz_t n;
  mpz_t pri_exp;
  mpz_t pub_exp;
  mpz_t payload;

  mpz_init_set_ui(n, (unsigned long)1517); //41 * 37
  mpz_init_set_ui(pri_exp, (unsigned long)857);
  mpz_init_set_ui(pub_exp, (unsigned long)1193);
  mpz_init_set_ui(payload, (unsigned long)42);

  mpz_t signed_payload;
  mpz_t verified_payload;

  mpz_t signed_payload_expected;
  mpz_t verified_payload_expected;

  mpz_init(signed_payload);
  mpz_init(verified_payload);

  mpz_init_set_ui(signed_payload_expected, (unsigned long)1108);
  mpz_init_set_ui(verified_payload_expected, (unsigned long)42);

  rsa_sign_payload(signed_payload, payload, pri_exp, n);
  rsa_verify_signature(verified_payload, signed_payload, pub_exp, n);
  
  int passes = 1;

  fprintf(stdout, "\n===== RSA TEST =====\n\n");

  gmp_fprintf(stdout, "=== Components ===\n\n");
  gmp_fprintf(stdout, "Modulus: %Zd\n", n);
  gmp_fprintf(stdout, "Pri Exp: %Zd\n", pri_exp);
  gmp_fprintf(stdout, "Pub Exp: %Zd\n", pub_exp);
  gmp_fprintf(stdout, "Payload: %Zd\n\n---\n\n", payload);

  gmp_fprintf(stdout, "Signed: %Zd\n", signed_payload);
  if (!mpz_cmp(signed_payload, signed_payload_expected)) {
    fprintf(stdout, "The signature is as expected.\n\n");
  } else {
    fprintf(stdout, "The signaute is not as expected.\n\n");
    passes = 0;
  }

  gmp_fprintf(stdout, "Verified: %Zd\n", verified_payload);
  if (!mpz_cmp(verified_payload, verified_payload_expected)) {
    fprintf(stdout, "The verification passed.\n\n");
  } else {
    fprintf(stdout, "The verification failed.\n\n");
    passes = 0;
  }

  if (passes) {
    fprintf(stdout, "=== RSA TEST: PASS ===\n\n");
  } else {
    fprintf(stdout, "=== RSA TEST: FAIL ===\n\n");
  }
}


int main() {
  rsa_test();
};
