/* 
 * client.c
 * a simple http client - code taken, with additions and alterations, from Davies' Implementing SSL/TLS Using Cryptography and PKI.
 * sam judson | cryptography and network protocols | marlboro college
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define HTTP_PORT 80



/* 
 *   Parse URL Input
 */



void parse_url(char* url, char** host, char** path) {
  char* host_pos;
  char* path_pos;

  host_pos = strstr(url, "//");
  
  if (host_pos) {
    *host = host_pos + 2;
  } else {
    *host = url;
  }

  path_pos = strchr(*host, '/');

  if (!path_pos) {
    *path = "/";
  } else {
    *path_pos = '\0';
    *path = path_pos + 1;
  }
}

void read_url(int argc, char* argv[], char** host, char** path) {
  int url_pos;

  for (url_pos = 0; url_pos < argc; url_pos++) {
    if (strcmp(argv[url_pos], "-u") == 0) {
      url_pos++; 
      break;
    } else if ((strcmp(argv[url_pos], "-h")) == 0) {
      url_pos = argc;
      break;
    }
  }

  if (url_pos == argc) {	     
    fprintf(stderr, "USAGE: client -u URL [-p post_payload]\nRequest Cap: 255 bytes\n");
    exit(-1);
  } else {
    parse_url(argv[url_pos], host, path);
  }
}



/*  
 *   Create Connection
 */



int establish_connection(struct hostent* host_name, struct sockaddr_in host_address, char* host, char* path) {
  int socket_id = socket(PF_INET, SOCK_STREAM, 0);

  if (socket_id == -1) {
    perror("ERROR: CODE 2 - UNABLE TO CREATE LOCAL SOCKET.\n");
    exit(2);
  }
  
  host_name = gethostbyname(host);

  if (host_name == -1) {
    perror("ERROR: CODE 3 - ERROR IN NAME RESOLUTION. PLEASE CONFIRM HOST.\n");
    exit(3);
  }

  host_address.sin_family = AF_INET;
  host_address.sin_port = htons(HTTP_PORT);
  memcpy(&host_address.sin_addr, host_name->h_addr_list[0], sizeof(struct in_addr));

  if (connect(socket_id, (struct sockaddr*) &host_address, sizeof(host_address)) == -1) {
    perror("ERROR: CODE 4 - UNABLE TO CONNECT TO HOST.\n");
    exit(4);
  }

  fprintf(stdout, "Connection established to host.\n");
  return socket_id;
}



/*
 *   Make HTTP Request
 */



void get(int connection, const char* host, const char* path) {
  static char get_request[255];
  
  if (path != "/") {
    sprintf(get_request, "GET /%s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n", path, host);
  } else {
    sprintf(get_request, "GET %s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n", path, host);
  }

  fprintf(stdout, "\n=========\n=REQUEST=\n=========\n\n%s", get_request);

  if (send(connection, get_request, strlen(get_request), 0) == -1) {
    perror("ERROR: CODE 5 - HTTP GET REQUEST FAILED.\n");
    exit(5);
  } 
}

void post(int connection, const char* host, const char* path, const char* payload) {
  static char post_request[255]; 
  
  if (path != "/") {
    sprintf(post_request, "POST /%s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n%s\r\n\r\n", path, host, payload);
  } else {
    sprintf(post_request, "POST %s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n%s\r\n\r\n", path, host, payload);
  }

  fprintf(stdout, "\n=========\n=REQUEST=\n=========\n\n%s", post_request);
  
  if (send(connection, post_request, strlen(post_request), 0) == -1) {
    perror("ERROR: CODE 6 - HTTP POST REQUEST FAILED.\n");
    exit(6);
  } 
}

void make_request(int argc, char* argv[], int connection, char* host, char* path){
  int payload_pos;

  for (payload_pos = 0; payload_pos < argc; payload_pos++) {
    if (strcmp(argv[payload_pos], "-p") == 0) {
      payload_pos++; 
      break;
    }
  }

  if (payload_pos == argc) {
    get(connection, host, path);
  } else {
    post(connection, host, path, argv[payload_pos]);
  }
}



/*
 *  Display Response
 */



void display_response(int connection) {
  int received = 0;
  static char recv_buf[255 + 1];

  fprintf(stdout, "==========\n=RESPONSE=\n==========\n\n");
                   
  while ((received = recv(connection, recv_buf, 255, 0)) > 0) {
    recv_buf[received] = '\0';
    fprintf(stdout, "%s", recv_buf);
  }

  fprintf(stdout, "\n");
}



/* 
 *   Main
 */



int main(int argc, char* argv[]) {
  int connection;
  char* host;
  char* path;
  struct hostent* host_name;
  struct sockaddr_in host_address;

  read_url(argc, argv, &host, &path);
  
  fprintf(stdout, "Specified host: %s\n", host);
  fprintf(stderr, "Specified path: %s\n", path);

  connection = establish_connection(host_name, host_address, host, path);
  make_request(argc, argv, connection, host, path);
  display_response(connection);

  return 0;
}
