/*
 * server.c
 * a simple server client - code taken, with additions and alterations, from Davies' Implementing SSL/TLS Using Cryptography and PKI.
 * sam judson | cryptography and network protocols | marlboro college
 */

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define HTTP_PORT 80
#define MAX_CONNECTIONS 16

#define LOGGING 1
#define LOG_FILE "server.log"



/*
 *  Create And Configure Listening Socket
 */



int gen_listen_socket() {
  int on = 1;
  int listen_socket_id = socket(PF_INET, SOCK_STREAM, 0);

  if (listen_socket_id == -1) {
    perror("ERROR: CODE 2 - UNABLE TO CREATE LISTENING SOCKET.\n");
    exit(2);
  }

  if (setsockopt(listen_socket_id, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1) {
    perror("ERROR: CODE 3 - UNABLE TO SET SOCKET OPTION.\n");
    exit(3);
  }

  return listen_socket_id;
}

void config_listen_socket(int listen_socket) {
  struct sockaddr_in local_addr;

  local_addr.sin_family = AF_INET;
  local_addr.sin_port = htons(HTTP_PORT);
  //local_addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  local_addr.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind(listen_socket, (struct sockaddr*) &local_addr, sizeof(local_addr)) == -1) {
    perror("ERROR: CODE 4 - UNABLE TO BIND TO LOCAL ADDRESS.\n");
    exit(4);
  }

  if (listen(listen_socket, 5) == -1) {
    perror("ERROR: CODE 5 - UNABLE TO CREATE BACKLOG.\n");
    exit(5);
  }
}



/*
 *  Request Handling Helpers
 */



struct connection_details {
  int conn_num;
  int* conn_sockets;
};

void log_server(int* conn_sockets, int conn_num, int opened) {
  if (LOGGING) {
    int i;

    FILE *fp;
    fp = fopen(LOG_FILE, "a");
    
    time_t rawtime;
    struct tm * timeinfo;

    // http://stackoverflow.com/a/5142028
    time(&rawtime);
    timeinfo = localtime(&rawtime);
    fprintf(fp, "%s - ", asctime(timeinfo));

    if (opened) {
      fprintf(fp, "New Connection Created. ID: %d\n", conn_num);
    } else {
      fprintf(fp, "Connection Closed. ID: %d\n", conn_num);
    }

    fprintf(fp, "Server Status: ");
    for (i = 0; i < MAX_CONNECTIONS; i++) {
      if (conn_sockets[i]) {
	fprintf(fp, "1");
      } else {
	fprintf(fp, "0");
      }
    }
    fprintf(fp, "\n\n");
    fclose(fp);
  }
}



/*
 *  Process HTTP Request
 */


char* read_line(int connection) {
  static int line_len = 255;
  static char* line = NULL;
  
  int size;
  char c;
  int pos = 0;

  if (!line) {
    line = malloc(line_len * sizeof(char)); //Just for good habit.
  }
  
  while ((size = recv(connection, &c, 1, 0)) > 0) {
    if ((c == '\n') && (line[pos - 1] == '\r')) {
      line[pos - 1] = '\0';
      break;
    }

    line[pos++] = c;
    
    if (pos > line_len) {
      line_len *= 2;
      line = realloc(line, line_len);
    } 
  }

  return line;
}

void handle_get(int connection) {
  char buf[255];
  char* request_line;
  int seen_empty_line = 0;
  
  while (!seen_empty_line) {
    request_line = read_line(connection);

    if (!strlen(request_line)) {
      seen_empty_line = 1;
    }
  }

  sprintf(buf, "HTTP/1.1 200 Success\r\nConnection: close\r\nContent-Type:text/html\r\n<html><head><title>Hello</title></head><body>Hello there, stranger.</body></html>\r\n");
  
  if (send(connection, buf, strlen(buf), 0) < strlen(buf)) {
    perror("ERROR: CODE 6 - UNABLE TO RESPOND.");
  }

}

void handle_post(int connection) {
  char* buf;
  char* request_line;
  char* name;
  char* name_end;
  int seen_payload = 0;

  while (!seen_payload) {
    request_line = read_line(connection);

    if (strchr(request_line, '=')) {
      seen_payload = 1;
      name = strstr(request_line, "name");
      
      if (name) {
	name = name + 5;
	name_end = strchr(name, '&');
	if (name_end) {
	  name_end = '\0';
	}
      } 
    }
  }

  if (name) {
    buf = malloc((255 + strlen(name)) * sizeof(char));
    sprintf(buf, "HTTP/1.1 200 Success\r\nConnection: close\r\nContent-Type:text/html\r\n<html><head><title>Hello</title></head><body>Hello there, %s.</body></html>\r\n", name);
  } else {
    buf = malloc(255 * sizeof(char));
    sprintf(buf, "HTTP/1.1 200 Success\r\nConnection: close\r\nContent-Type:text/html\r\n<html><head><title>Hello</title></head><body>Hello there, partner.</body></html>\r\n", name);
  }

  if (send(connection, buf, strlen(buf), 0) < strlen(buf)) {
    perror("ERROR: CODE 6 - UNABLE TO RESPOND.");
  }

  free(buf);
}

void handle_other(int connection) {
  char buf[255];
  char* request_line;
  int seen_empty_line = 0;
  
  while (!seen_empty_line) {
    request_line = read_line(connection);

    if (!strlen(request_line)) {
      seen_empty_line = 1;
    }
  }

  sprintf(buf, "HTTP/1.1 501 Error Occurred\r\n\r\n");
  
  if (send(connection, buf, strlen(buf), 0) < strlen(buf)) {
    perror("ERROR: CODE 6 - UNABLE TO RESPOND.");
  }
}

void* process_request(void* conn_d_ptr) {
  struct connection_details conn_d = *((struct connection_details *) conn_d_ptr);
  
  char* request_line;
  request_line = read_line(conn_d.conn_sockets[conn_d.conn_num]);
  if (!strncmp(request_line, "GET", 3)) {
    handle_get(conn_d.conn_sockets[conn_d.conn_num]);
  } else if (!strncmp(request_line, "POST", 4)) {
    handle_post(conn_d.conn_sockets[conn_d.conn_num]);
  } else {
    handle_other(conn_d.conn_sockets[conn_d.conn_num]);
  }

  shutdown(conn_d.conn_sockets[conn_d.conn_num], 2);
  log_server(conn_d.conn_sockets, conn_d.conn_num, 0);
  pthread_exit(NULL);
}



/*
 *  Accept Connections
 */



void zero_connection_array(int conn_sockets[MAX_CONNECTIONS]) {
  int i;

  for (i=0; i < MAX_CONNECTIONS; i++) {
    conn_sockets[i] = 0;
  }
}

int next_open_conn(int conn_sockets[MAX_CONNECTIONS]) {
  int i;

  for (i = 0; i < MAX_CONNECTIONS; i++) {
    if (!conn_sockets[i]) {
      return i;
    }
  } 
  return -1;
}

void accept_connections(int listen_socket) {
  int temp_connect_socket, conn_num;

  int conn_sockets[MAX_CONNECTIONS]; // by default supports sixteen threads (connections) concurrently
  pthread_t threads[MAX_CONNECTIONS]; 

  zero_connection_array(conn_sockets);

  struct sockaddr_in client_addr;
  int client_addr_len = sizeof(client_addr);

  struct connection_details conn_d;

  fprintf(stderr, "Listening on port: %d", HTTP_PORT);

  while ((temp_connect_socket = accept(listen_socket, (struct sockaddr*) &client_addr, &client_addr_len)) != -1) {
    if ((conn_num = next_open_conn(conn_sockets)) != -1) {
      conn_sockets[conn_num] = temp_connect_socket; 

      conn_d.conn_num = conn_num;
      conn_d.conn_sockets = conn_sockets;

      pthread_create(&threads[conn_num], NULL, process_request, (void *)&conn_d);
      log_server(conn_sockets, conn_num, 1);
    }
  }
}



/*
 *  MAIN
 */



int main(int argc, char* argv[]) {
  int listen_socket;
  
  listen_socket = gen_listen_socket();
  config_listen_socket(listen_socket);

  accept_connections(listen_socket);
  
  return 0;
}
