/*
 * serpent.c
 * an implementation of the serpent block cipher
 * samuel judson | cryptography and network protocols | marlboro college 
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "serpent.h"
#include "serpent_lib/encrypt.h"
#include "serpent_lib/decrypt.h"

#ifndef SPECS
#include "serpent_lib/cipher_specs.h"
#define SPECS 1
#endif

#define ENC 1
#define DEC 2
#define T 1
#define F 0



/*
 *  Get Or Generate Key
 */



void read_keyfile(char* path, uint32_t key[KEY_LENGTH]) {
  int i;

  FILE *fp;
  fp = fopen(path, "r");

  for (i = 0; i < KEY_LENGTH; i++) {
    fscanf(fp, "%08X", &key[i]);
  }
}

void export_key(uint32_t key[KEY_LENGTH]) {
  int i;

  FILE *fp;
  fp = fopen("key.out", "wb");
  
  for (i = 0; i < KEY_LENGTH; i++) {
    fprintf(fp, "%08X", key[i]);
  }
}

void generate_key(uint32_t key[KEY_LENGTH]) {
  int i, rand_num;

  FILE* urand;
  urand = fopen("/dev/urandom", "rb");

  for (i = 0; i < KEY_LENGTH; i++) {
    fread(&rand_num, sizeof(uint32_t), 1, urand);
    key[i] = (uint32_t)rand_num;
  }

  export_key(key);
}



/*
 *  Print Help Dialog
 */



void print_help() {
  fprintf(stdout, "Usage: serpent [MODE] [KEY]... [INPUT]...\n");
  fprintf(stdout, "Symmetric key encryption using the Serpent cipher.\n\n");
  fprintf(stdout, "Requires one of -e/-d, one of -f/-g (latter only valid for encryption), and -i\n\n");
  fprintf(stdout, "  -e, --encrypt      encrypt the input file\n");
  fprintf(stdout, "  -d, --decrypt      decrypt the input file\n");
  fprintf(stdout, "  -f, --key-file     the path to a file containing a key\n");
  fprintf(stdout, "  -g, --gen-key      generate a key\n");
  fprintf(stdout, "  -i, --input-file   the file containing the input to be en/decrypted\n");
  fprintf(stdout, "  -h, --help         print this dialog\n\n");

  exit(-1);
} 



/*
 *  Parse Arguments
 */



arguments parse_args(int argc, char* argv[]) {
  int i;
  arguments args;

  args.mode = 0;

  for (i = 0; i < argc; i++) {

    if (!strcmp(argv[i], "-e") || !strcmp(argv[i], "--encrypt")) {
      args.mode = ENC;

    } else if (!strcmp(argv[i], "-d") || !strcmp(argv[i], "--decrypt")) {
      args.mode = DEC;

    } else if (!strcmp(argv[i], "-g") || !strcmp(argv[i], "--gen-key")) {
      generate_key(args.key);

    } else if (!strcmp(argv[i], "-f") || !strcmp(argv[i], "--key-file")) {
      read_keyfile(argv[i+1], args.key);

    } else if (!strcmp(argv[i], "-i") || !strcmp(argv[i], "--input-file")) {
      strcpy(args.input_file, argv[i+1]);

    } else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
      print_help();
    }
  }

  if (!args.mode || !args.key || !args.input_file) {
    print_help();
  } else {
    return args; 
  }
}



/*
 *  File I/O
 */



char* read_input_file(char* input_file) {
  char* input;
  int size;

  FILE* fp;
  fp = fopen(input_file, "rb");
  
  if (!fp) {
    perror("Error: Invalid file path.\n");
    exit(-1);
  }

  fseek(fp, 0, SEEK_END);
  size = ftell(fp);
  fseek(fp, 0, SEEK_SET);

  input = malloc((size + 1) * sizeof(char));
  fscanf(fp, "%s", input);

  fclose(fp);

  return input;
}

void write_output(int type, char* output) {
  FILE* fp;

  if (type == ENC) {
    fp = fopen("serpent.enc", "wb");
  } else {
    fp = fopen("serpent.dec", "wb");
  }

  fprintf(fp, "%s", output);
}


/*
 *  Encrypt/Decrypt Wrappers
 */



char* encrypt(uint32_t key[KEY_LENGTH], char* input) {
  return serpent_encrypt(key, input, F);
}

char* decrypt(uint32_t key[KEY_LENGTH], char* input) {
  return serpent_decrypt(key, input, F);
}



/*
 *  MAIN
 */



int main(int argc, char* argv[]) {
  arguments args;
  char* input;
  char* plaintext;
  char* ciphertext;
  
  args = parse_args(argc, argv);

  input = read_input_file(args.input_file);

  if (args.mode == ENC) {
    ciphertext = encrypt(args.key, input);
    write_output(ENC, ciphertext);
  } else {
    plaintext = decrypt(args.key, input);
    write_output(DEC, plaintext);
  }

  return 0;
}
