/*
 *  serpent_lib/encrypt.c
 *  encrypts a given plaintext file with a given key
 *  samuel judson | cryptography and network protocols | marlboro college
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "pad.h"
#include "key_schedule.h"

#ifndef SBOXES
#include "sboxes.h"
#define SBOXES 1
#endif

#ifndef SPECS
#include "cipher_specs.h"
#define SPECS 1
#endif

#define T 1
#define F 0


/*
 *  CBC
 */


void decrypt_cbc(uint32_t current_block[BLOCK_LENGTH], uint32_t past_block[BLOCK_LENGTH]) {
  int i;

  for (i = 0; i < BLOCK_LENGTH; i++) {
    current_block[i] ^= past_block[i];
  }
}



/*
 *  Decrypt Block
 */




void decrypt_export_block(char* output, uint32_t current_block[BLOCK_LENGTH], int last_byte_loc) {
  int i, j, last_byte_chunk, last_byte_placement, curr_out_size;
  unsigned char output_byte;
  char chunk_output[BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE + 1];

  if (last_byte_loc < (BYTE_CHUNK_SIZE * 1)) {
    last_byte_chunk = 0;
  } else if (last_byte_loc >= (BYTE_CHUNK_SIZE * 1) && last_byte_loc < (BYTE_CHUNK_SIZE * 2)) {
    last_byte_chunk = 1;
  } else if (last_byte_loc >= (BYTE_CHUNK_SIZE * 2) && last_byte_loc < (BYTE_CHUNK_SIZE * 3)) {
    last_byte_chunk = 2;
  } else if (last_byte_loc >= (BYTE_CHUNK_SIZE * 3) && last_byte_loc < (BYTE_CHUNK_SIZE * 4)) {
    last_byte_chunk = 3;
  } else {
    last_byte_chunk = 4;
  }

  curr_out_size = strlen(output);
  output = (char*) realloc(output, curr_out_size + (last_byte_chunk * BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE));

  for (i = 0; i < last_byte_chunk; i++) {
    if (i < last_byte_chunk) {
      sprintf(chunk_output, "%08X", current_block[i]);
      strcat(output, chunk_output);
    } else {
      last_byte_placement = last_byte_loc % BYTE_CHUNK_SIZE; 

      for (j = 0; j < last_byte_placement; j++) {
	memcpy(&output_byte, (char*)&current_block[i] + (j * BYTE), BYTE);
	sprintf(output, "%02X", output_byte);
	strcat(output, chunk_output);
      }
    }
  }
}

void decrypt_block(uint32_t round_keys[NUM_ROUND_KEYS][ROUND_KEY_LENGTH], uint32_t current_block[BLOCK_LENGTH], uint32_t past_block[BLOCK_LENGTH], char* output, int test) {
  int last_byte_loc;
  uint32_t next_past_block[BLOCK_LENGTH]; // stores next_block so it can be set to current_block after CBC after decryption

  memcpy(next_past_block, current_block, BLOCK_LENGTH * BYTE_CHUNK_SIZE);

  /*
   *  START CODE BLOCK
   *
   *  Following code from serpent.c file in official submission code (http://www.cl.cam.ac.uk/~rja14/Papers/serpent.tar.gz) with minor name changes. See 'sboxes.h' for note on copyright.
   */

  register uint32_t x0, x1, x2, x3;
  register uint32_t y0, y1, y2, y3;

  x0 = current_block[0];
  x1 = current_block[1];
  x2 = current_block[2];
  x3 = current_block[3];

  /* Start to decrypt the ciphertext x */
  keying(x0, x1, x2, x3, round_keys[32]);
  InvRND31(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[31]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND30(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[30]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND29(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[29]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND28(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[28]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND27(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[27]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND26(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[26]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND25(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[25]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND24(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[24]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND23(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[23]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND22(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[22]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND21(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[21]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND20(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[20]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND19(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[19]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND18(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[18]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND17(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[17]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND16(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[16]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND15(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[15]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND14(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[14]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND13(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[13]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND12(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[12]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND11(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[11]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND10(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[10]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND09(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[ 9]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND08(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[ 8]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND07(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[ 7]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND06(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[ 6]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND05(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[ 5]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND04(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[ 4]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND03(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[ 3]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND02(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[ 2]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND01(x0, x1, x2, x3, y0, y1, y2, y3);
  keying(y0, y1, y2, y3, round_keys[ 1]);
  inv_transform(y0, y1, y2, y3, x0, x1, x2, x3);
  InvRND00(x0, x1, x2, x3, y0, y1, y2, y3);
  x0 = y0; x1 = y1; x2 = y2; x3 = y3;
  keying(x0, x1, x2, x3, round_keys[ 0]);
  /* The plaintext is now in x */

  current_block[0] = x0;
  current_block[1] = x1;
  current_block[2] = x2;
  current_block[3] = x3;
  
  /*
   *  END CODE BLOCK
   */

  if (!test) {
    decrypt_cbc(current_block, past_block);
  }

  last_byte_loc = unpad(current_block);

  decrypt_export_block(output, current_block, last_byte_loc);
  memcpy(past_block, next_past_block, BLOCK_LENGTH * BYTE_CHUNK_SIZE);
}



/*
 *  Decrypt Plaintext
 */



void read_iv(char** input_ptr,  uint32_t iv[BLOCK_LENGTH]) {
  int i;

  for (i = 0; i < BLOCK_LENGTH; i++) {
    sscanf(*(input_ptr), "%08X", &iv[i]); 
    *(input_ptr) = *(input_ptr) + BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE;
  } 
}

int decrypt_read_block(char** input_ptr, uint32_t next_block[BLOCK_LENGTH]) {
  int i, read, done;
  unsigned char next_char;

  for (i = 0; i < BLOCK_LENGTH; i++) {
    sscanf(*(input_ptr), "%08X", &next_block[i]); 
    *(input_ptr) = *(input_ptr) + BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE;
  }
  
  if (strlen(*(input_ptr))) {
    done = F;
  } else {
    done = T;
  }

  return done;
}

char* serpent_decrypt(uint32_t key[KEY_LENGTH], char* input, int test) {
  char* output;
  output = (char*) malloc(sizeof(char));
  output[0] = '\0';

  int done = F;
  uint32_t round_keys[NUM_ROUND_KEYS][ROUND_KEY_LENGTH];  uint32_t current_block[BLOCK_LENGTH];
  uint32_t past_block[BLOCK_LENGTH];
  uint32_t iv[BLOCK_LENGTH];
  char** input_ptr = &input;
  
  key_schedule(round_keys, key);

  if (!test) {
    read_iv(input_ptr, iv); /* If test there isn't an IV */
  } 
  
  memcpy(past_block, iv, BLOCK_LENGTH * BYTE_CHUNK_SIZE);

  while (!done) {
    done = decrypt_read_block(input_ptr, current_block); 

    decrypt_block(round_keys, current_block, past_block, output, test); // the function sets current_block from this call to past_block for next
  }

  return output;
}
