/*
 *  serpent_lib/encrypt.c
 *  encrypts a given plaintext file with a given key
 *  samuel judson | cryptography and network protocols | marlboro college
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "pad.h"
#include "key_schedule.h"

#ifndef SBOXES
#include "sboxes.h"
#define SBOXES 1
#endif

#ifndef SPECS
#include "cipher_specs.h"
#define SPECS 1
#endif

#define T 1
#define F 0



/*
 *  CBC
 */



void encrypt_cbc(uint32_t current_block[BLOCK_LENGTH], uint32_t past_block[BLOCK_LENGTH]) {
  int i;
 
  for (i = 0; i < BLOCK_LENGTH; i++) {
    current_block[i] ^= past_block[i];
  }
}



/*
 *  Encrypt Block
 */


void encrypt_export_block(char* output, uint32_t current_block[BLOCK_LENGTH]) {
  int i, curr_out_size;
  char chunk_output[BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE + 1];

  curr_out_size = strlen(output);
  output = (char*) realloc(output, curr_out_size + (BLOCK_LENGTH * BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE));

  for (i = 0; i < BLOCK_LENGTH; i++) {
    sprintf(chunk_output, "%08X", current_block[i]);
    strcat(output, chunk_output);
  }
}

void encrypt_block(uint32_t round_keys[NUM_ROUND_KEYS][ROUND_KEY_LENGTH], uint32_t current_block[BLOCK_LENGTH], uint32_t past_block[BLOCK_LENGTH], char* output, int test) {  
  
  if (!test) {
    encrypt_cbc(current_block, past_block);
  }

  /*
   *  START CODE BLOCK
   *
   *  Following code from serpent.c file in official submission code (http://www.cl.cam.ac.uk/~rja14/Papers/serpent.tar.gz) with minor name changes. See 'sboxes.h' for note on copyright.
   */

  register uint32_t x0, x1, x2, x3;
  register uint32_t y0, y1, y2, y3;

  x0 = current_block[0];
  x1 = current_block[1];
  x2 = current_block[2];
  x3 = current_block[3];

  /* Start to encrypt the plaintext x */
  keying(x0, x1, x2, x3, round_keys[ 0]);
  RND00(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[ 1]);
  RND01(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[ 2]);
  RND02(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[ 3]);
  RND03(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[ 4]);
  RND04(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[ 5]);
  RND05(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[ 6]);
  RND06(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[ 7]);
  RND07(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[ 8]);
  RND08(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[ 9]);
  RND09(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[10]);
  RND10(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[11]);
  RND11(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[12]);
  RND12(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[13]);
  RND13(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[14]);
  RND14(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[15]);
  RND15(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[16]);
  RND16(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[17]);
  RND17(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[18]);
  RND18(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[19]);
  RND19(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[20]);
  RND20(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[21]);
  RND21(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[22]);
  RND22(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[23]);
  RND23(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[24]);
  RND24(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[25]);
  RND25(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[26]);
  RND26(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[27]);
  RND27(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[28]);
  RND28(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[29]);
  RND29(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[30]);
  RND30(x0, x1, x2, x3, y0, y1, y2, y3);
  transform(y0, y1, y2, y3, x0, x1, x2, x3);
  keying(x0, x1, x2, x3, round_keys[31]);
  RND31(x0, x1, x2, x3, y0, y1, y2, y3);
  x0 = y0; x1 = y1; x2 = y2; x3 = y3;
  keying(x0, x1, x2, x3, round_keys[32]);
  /* The ciphertext is now in x */

  current_block[0] = x0;
  current_block[1] = x1;
  current_block[2] = x2;
  current_block[3] = x3;
  
  /*
   *  END CODE BLOCK
   */

  encrypt_export_block(output, current_block);
  memcpy(past_block, current_block, BLOCK_LENGTH * BYTE_CHUNK_SIZE);
}



/*
 *  Encrypt Plaintext
 */



int encrypt_read_block(char** input_ptr, uint32_t current_block[BLOCK_LENGTH]) {
  int i, j, bytes_to_pad;
  int done = F;
  unsigned char current_byte[HEX_CHARS_PER_BYTE];
  unsigned char current_chunk[BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE];

  for (i = 0; i < BLOCK_LENGTH; i++) {
    for (j = 0; j < BYTE_CHUNK_SIZE; j++) {
      
      if (strlen(*(input_ptr))) {
	sscanf(*(input_ptr), "%2c", current_byte);
	memcpy(&current_chunk[(j * HEX_CHARS_PER_BYTE)], &current_byte, HEX_CHARS_PER_BYTE);
	*(input_ptr) = *(input_ptr) + HEX_CHARS_PER_BYTE;
      } else {
	if (!j) {
	  i--;
	}
	done = T;
	break;
      }
    }
   
    if (j) {
      sscanf(current_chunk, "%08X", &current_block[i]); 
    }

    if (done) {
      if (!(i > 0) && !(j > 0)) {
	done = -1;
	break;
      } else {
	bytes_to_pad = (BLOCK_LENGTH * BYTE_CHUNK_SIZE) - ((BLOCK_LENGTH * i) + BYTE_CHUNK_SIZE - j);
	pad(current_block, bytes_to_pad);
	break;
      }
    }
  }

  //exit(-1);

  return done;
}

void generate_iv(uint32_t iv[BLOCK_LENGTH], char* output) {
  int i, rand_num;

  FILE* urand;
  urand = fopen("/dev/urandom", "rb");

  for (i = 0; i < BLOCK_LENGTH; i++) {
    fread(&rand_num, sizeof(uint32_t), 1, urand);
    iv[i] = (uint32_t)rand_num;
  }
  
  encrypt_export_block(output, iv);
}

// For testing no chaining (and therefore IV generation/output) is done.
char* serpent_encrypt(uint32_t key[KEY_LENGTH], char* input, int test) {
  char* output;
  output = (char*) malloc(sizeof(char));
  output[0] = '\0';

  uint32_t round_keys[NUM_ROUND_KEYS][ROUND_KEY_LENGTH];
  uint32_t current_block[BLOCK_LENGTH];
  uint32_t past_block[BLOCK_LENGTH];
  uint32_t iv[BLOCK_LENGTH];
  int done = 0;
  char** input_ptr = &input;
  
  key_schedule(round_keys, key);

  if (!test) {
    generate_iv(iv, output);
  }
  
  memcpy(past_block, iv, BLOCK_LENGTH * BYTE_CHUNK_SIZE);

  while (!done) {

    /* done is 1 if there is no more data but there is some in the current block to encrypt
               0 if the current block is full with data to encrypt
	      -1 if there is no more data and none in the current block to encrypt             */
    done = encrypt_read_block(input_ptr, current_block);
    
    if (done >= 0) { 
      encrypt_block(round_keys, current_block, past_block, output, test); // the function sets current_block from this call to past_block for next
    }
  }

  return output;
}
