/*
 *  serpent_lib/key_schedule.c
 *  creates the round keys from the main key for serpent-256
 *  samuel judson | cryptography and network protocols | marlboro college
 */

#include <stdio.h>
#include <string.h>

#include "sboxes.h"
#include "cipher_specs.h"

#define PHI 0x9e3779b9L


/*
 *  Create Pre- and Intermediate Keys
 */



void get_prekeys(uint32_t prekeys[NUM_PREKEYS], uint32_t key[KEY_LENGTH]) {		   
  int i;

  for (i = 0; i < NUM_PREKEYS; i++) {
    prekeys[i] = key[i];
  }
}

void prekey_to_intermediate(uint32_t intermediate_keys[NUM_INTERMEDIATE_KEYS], uint32_t prekeys[NUM_PREKEYS]) {
  int i, j;

  for (i = 0; i < NUM_PREKEYS; i++) {
    intermediate_keys[i] = prekeys[i];
  } 

  for (i = 8; i < 132; i++) {
    intermediate_keys[i] = ROL(intermediate_keys[i-8] ^ intermediate_keys[i-5] ^ intermediate_keys[i-3] ^ intermediate_keys[i-1] ^ PHI ^ i, 11);
  }
}



/*
 *  Create Round Keys
 */



void intermediate_to_round(uint32_t round_keys[NUM_ROUND_KEYS][ROUND_KEY_LENGTH], uint32_t intermediate_keys[NUM_INTERMEDIATE_KEYS]) {
  int i, j;
  uint32_t sbox_output[NUM_INTERMEDIATE_KEYS];

  /*
   *  START CODE BLOCK
   *
   *  Following code from serpent.c file in official submission code (http://www.cl.cam.ac.uk/~rja14/Papers/serpent.tar.gz) with minor name changes. See 'sboxes.h' for note on copyright.
   */

  RND03(intermediate_keys[0], intermediate_keys[1], intermediate_keys[2], intermediate_keys[3], sbox_output[0], sbox_output[1], sbox_output[2], sbox_output[3]);
  RND02(intermediate_keys[4], intermediate_keys[5], intermediate_keys[6], intermediate_keys[7], sbox_output[4], sbox_output[5], sbox_output[6], sbox_output[7]);
  RND01(intermediate_keys[8], intermediate_keys[9], intermediate_keys[10], intermediate_keys[11], sbox_output[8], sbox_output[9], sbox_output[10], sbox_output[11]);
  RND00(intermediate_keys[12], intermediate_keys[13], intermediate_keys[14], intermediate_keys[15], sbox_output[12], sbox_output[13], sbox_output[14], sbox_output[15]);
  RND31(intermediate_keys[16], intermediate_keys[17], intermediate_keys[18], intermediate_keys[19], sbox_output[16], sbox_output[17], sbox_output[18], sbox_output[19]);
  RND30(intermediate_keys[20], intermediate_keys[21], intermediate_keys[22], intermediate_keys[23], sbox_output[20], sbox_output[21], sbox_output[22], sbox_output[23]);
  RND29(intermediate_keys[24], intermediate_keys[25], intermediate_keys[26], intermediate_keys[27], sbox_output[24], sbox_output[25], sbox_output[26], sbox_output[27]);
  RND28(intermediate_keys[28], intermediate_keys[29], intermediate_keys[30], intermediate_keys[31], sbox_output[28], sbox_output[29], sbox_output[30], sbox_output[31]);
  RND27(intermediate_keys[32], intermediate_keys[33], intermediate_keys[34], intermediate_keys[35], sbox_output[32], sbox_output[33], sbox_output[34], sbox_output[35]);
  RND26(intermediate_keys[36], intermediate_keys[37], intermediate_keys[38], intermediate_keys[39], sbox_output[36], sbox_output[37], sbox_output[38], sbox_output[39]);
  RND25(intermediate_keys[40], intermediate_keys[41], intermediate_keys[42], intermediate_keys[43], sbox_output[40], sbox_output[41], sbox_output[42], sbox_output[43]);
  RND24(intermediate_keys[44], intermediate_keys[45], intermediate_keys[46], intermediate_keys[47], sbox_output[44], sbox_output[45], sbox_output[46], sbox_output[47]);
  RND23(intermediate_keys[48], intermediate_keys[49], intermediate_keys[50], intermediate_keys[51], sbox_output[48], sbox_output[49], sbox_output[50], sbox_output[51]);
  RND22(intermediate_keys[52], intermediate_keys[53], intermediate_keys[54], intermediate_keys[55], sbox_output[52], sbox_output[53], sbox_output[54], sbox_output[55]);
  RND21(intermediate_keys[56], intermediate_keys[57], intermediate_keys[58], intermediate_keys[59], sbox_output[56], sbox_output[57], sbox_output[58], sbox_output[59]);
  RND20(intermediate_keys[60], intermediate_keys[61], intermediate_keys[62], intermediate_keys[63], sbox_output[60], sbox_output[61], sbox_output[62], sbox_output[63]);
  RND19(intermediate_keys[64], intermediate_keys[65], intermediate_keys[66], intermediate_keys[67], sbox_output[64], sbox_output[65], sbox_output[66], sbox_output[67]);
  RND18(intermediate_keys[68], intermediate_keys[69], intermediate_keys[70], intermediate_keys[71], sbox_output[68], sbox_output[69], sbox_output[70], sbox_output[71]);
  RND17(intermediate_keys[72], intermediate_keys[73], intermediate_keys[74], intermediate_keys[75], sbox_output[72], sbox_output[73], sbox_output[74], sbox_output[75]);
  RND16(intermediate_keys[76], intermediate_keys[77], intermediate_keys[78], intermediate_keys[79], sbox_output[76], sbox_output[77], sbox_output[78], sbox_output[79]);
  RND15(intermediate_keys[80], intermediate_keys[81], intermediate_keys[82], intermediate_keys[83], sbox_output[80], sbox_output[81], sbox_output[82], sbox_output[83]);
  RND14(intermediate_keys[84], intermediate_keys[85], intermediate_keys[86], intermediate_keys[87], sbox_output[84], sbox_output[85], sbox_output[86], sbox_output[87]);
  RND13(intermediate_keys[88], intermediate_keys[89], intermediate_keys[90], intermediate_keys[91], sbox_output[88], sbox_output[89], sbox_output[90], sbox_output[91]);
  RND12(intermediate_keys[92], intermediate_keys[93], intermediate_keys[94], intermediate_keys[95], sbox_output[92], sbox_output[93], sbox_output[94], sbox_output[95]);
  RND11(intermediate_keys[96], intermediate_keys[97], intermediate_keys[98], intermediate_keys[99], sbox_output[96], sbox_output[97], sbox_output[98], sbox_output[99]);
  RND10(intermediate_keys[100], intermediate_keys[101], intermediate_keys[102], intermediate_keys[103], sbox_output[100], sbox_output[101], sbox_output[102], sbox_output[103]);
  RND09(intermediate_keys[104], intermediate_keys[105], intermediate_keys[106], intermediate_keys[107], sbox_output[104], sbox_output[105], sbox_output[106], sbox_output[107]);
  RND08(intermediate_keys[108], intermediate_keys[109], intermediate_keys[110], intermediate_keys[111], sbox_output[108], sbox_output[109], sbox_output[110], sbox_output[111]);
  RND07(intermediate_keys[112], intermediate_keys[113], intermediate_keys[114], intermediate_keys[115], sbox_output[112], sbox_output[113], sbox_output[114], sbox_output[115]);
  RND06(intermediate_keys[116], intermediate_keys[117], intermediate_keys[118], intermediate_keys[119], sbox_output[116], sbox_output[117], sbox_output[118], sbox_output[119]);
  RND05(intermediate_keys[120], intermediate_keys[121], intermediate_keys[122], intermediate_keys[123], sbox_output[120], sbox_output[121], sbox_output[122], sbox_output[123]);
  RND04(intermediate_keys[124], intermediate_keys[125], intermediate_keys[126], intermediate_keys[127], sbox_output[124], sbox_output[125], sbox_output[126], sbox_output[127]);
  RND03(intermediate_keys[128], intermediate_keys[129], intermediate_keys[130], intermediate_keys[131], sbox_output[128], sbox_output[129], sbox_output[130], sbox_output[131]);

  /*
   *  END CODE BLOCK
   */

  for (i = 0; i < NUM_ROUND_KEYS; i++) {
    for (j = 0; j < ROUND_KEY_LENGTH; j++) {
      round_keys[i][j] = sbox_output[(i * ROUND_KEY_LENGTH) + j];
    }
  } 
}



/*
 *  Create Round Keys Via Key Schedule
 */



void key_schedule(uint32_t round_keys[NUM_ROUND_KEYS][ROUND_KEY_LENGTH], uint32_t key[KEY_LENGTH]) {

  uint32_t prekeys[NUM_PREKEYS];
  uint32_t intermediate_keys[NUM_INTERMEDIATE_KEYS];

  get_prekeys(prekeys, key);
  prekey_to_intermediate(intermediate_keys, prekeys);
  intermediate_to_round(round_keys, intermediate_keys);
}
