/*
 * serpent_lib/pad.c
 * pad/unpad plaintexts for use in serpent cipher (or any block cipher with 128-bit block size)
 * sam judson | cryptography and network protocols | marlboro college
 */

#include <string.h>
#include <stdio.h>

#include "cipher_specs.h"



/*
 *  Unpad Block
 */ 



int unpad(uint32_t block[BLOCK_LENGTH]) {
  int last_byte_loc;
  int i, j;
  unsigned char padding_byte;
  unsigned char current_byte;

  memcpy(&padding_byte, (char*)&(block[BLOCK_LENGTH - 1]), BYTE); 

  if (padding_byte != 255 && padding_byte != 0) {
    return BLOCK_LENGTH * BYTE_CHUNK_SIZE; //Block isn't padded.
  }

  for (i = (BLOCK_LENGTH - 1); i >= 0; i--) {
    for (j = 0; j < BYTE_CHUNK_SIZE; j++) {  //Run forward to count down due to little endian-ness
      memcpy(&current_byte, (char*)&(block[i]) + (BYTE * j), BYTE);
      if (current_byte ^ padding_byte) {
	return (i * BYTE_CHUNK_SIZE) + (BYTE_CHUNK_SIZE - j);
      }
    }
  }
}



/*
 *  Pad Block
 */ 



void pad(uint32_t block[BLOCK_LENGTH], int bytes_to_pad) {
  unsigned char last_byte;
  unsigned char padding_byte;

  int i, j, k;
  int last_byte_loc = (BLOCK_LENGTH * BYTE_CHUNK_SIZE) - bytes_to_pad - 1;  // Last byte location. Subtract an additional one so bytes are ordered 0-15.
  int last_byte_chunk;                                                      // Chunk containing final byte.
  int last_byte_placement;                                                  // Byte's placement within chunk.

  uint32_t partial_chunk;
  int shift, start_loc, end_loc;
                                                     
  if (last_byte_loc < (BYTE_CHUNK_SIZE * 1)) {
    last_byte_chunk = 0;
  } else if (last_byte_loc >= (BYTE_CHUNK_SIZE * 1) && last_byte_loc < (BYTE_CHUNK_SIZE * 2)) {
    last_byte_chunk = 1;
  } else if (last_byte_loc >= (BYTE_CHUNK_SIZE * 2) && last_byte_loc < (BYTE_CHUNK_SIZE * 3)) {
    last_byte_chunk = 2;
  } else if (last_byte_loc >= (BYTE_CHUNK_SIZE * 3) && last_byte_loc < (BYTE_CHUNK_SIZE * 4)) {
    last_byte_chunk = 3;
  }
  
  last_byte_placement = last_byte_loc % BYTE_CHUNK_SIZE;  

  memcpy(&last_byte, (char*)&(block[last_byte_chunk]) + (last_byte_placement * BYTE), BYTE);

  if (last_byte & 1) {
    padding_byte = 0;
  } else {
    padding_byte = 0xFF;
  }

  // Pad rest of chunk containing last byte

  for (i = (last_byte_placement + 1); i < BYTE_CHUNK_SIZE; i++) {
    memcpy((char*)&(block[last_byte_chunk]) + (i * BYTE), &padding_byte, BYTE);
  }

  // Move padded bytes to front of chunk due to little endian-ness

  partial_chunk = block[last_byte_chunk];
  shift = (BYTE_CHUNK_SIZE - 1) - last_byte_placement; //Subtraction to account for byte numbering 0-3.

  for (j = 0; j < BYTE_CHUNK_SIZE; j++) {
    start_loc = j * BYTE;
    end_loc = ((j * BYTE) + shift) % BYTE_CHUNK_SIZE;
    memcpy((char*)&block[last_byte_chunk] + end_loc, (char*)&partial_chunk + start_loc, BYTE);
  }

  // Pad full chunks.

  for (k = (last_byte_chunk + 1); k < BLOCK_LENGTH; k++) {
    if (!padding_byte) {
      block[k] = 0;
    } else {
      block[k] = UINT32_MAX;
    }
  }
} 
