/*
 *  keygen/test.c
 *  tests the key generation functions for RSA and DHE
 *  samuel judson | cryptography and network protocols | marlboro college
 */

#include <stdio.h>
#include <string.h>
#include <gmp.h>

#include "keygen.h"

#ifndef RSA_SPECS
#include "../rsa/rsa_specs.h"
#define RSA_SPECS 1
#endif

#ifndef DHE_SPECS
#include "../dhe/dhe_specs.h"
#define dhe_SPECS 1
#endif

#define HEX_CHAR_PER_BYTE 2


void rsa_keygen_test() {
  char n_str[(RSA_MODULUS_LENGTH * HEX_CHAR_PER_BYTE) + 1];
  char pri_str[(RSA_MODULUS_LENGTH * HEX_CHAR_PER_BYTE) + 1];
  char pub_str[8];

  mpz_t n, pri_exp, pub_exp;
  mpz_init(n);
  mpz_init(pri_exp);
  mpz_init(pub_exp);

  int passes = 1;
  
  rsa_keygen(n, pri_exp, pub_exp);

  fprintf(stdout, "\n===== RSA KEY GENERATION TESTS (ALL OUTPUTS IN HEX)=====\n\n");

  mpz_get_str(n_str, 16, n);
  mpz_get_str(pri_str, 16, pri_exp);
  mpz_get_str(pub_str, 16, pub_exp);
  
  fprintf(stdout, "=== Generated Components ===\n\n");
  fprintf(stdout, "Modulus: %s\n\n", n_str);
  if (strlen(n_str) == RSA_MODULUS_LENGTH * HEX_CHAR_PER_BYTE) {
    fprintf(stdout, "Modulus has correct length.\n\n---\n\n");
  } else {
    passes = 0;
    fprintf(stdout, "Modulus has incorrect length.\n\n---\n\n");
  }

  fprintf(stdout, "Private Key Exp: %s\n\n", pri_str);
  fprintf(stdout, "Public Key Exp: %s\n\n", pub_str);

  if (passes) {
    fprintf(stdout, "\n=== RSA KEY GENERATION TEST: PASS ===\n\n");
  } else {
    fprintf(stdout, "\n=== RSA KEY GENERATION TEST: FAIL ===\n\n");
  }
}

void dhe_keygen_test() {
  char sec_str[DHE_KEY_LENGTH * HEX_CHAR_PER_BYTE];

  mpz_t rand_sec;
  mpz_init(rand_sec);

  int passes = 1;

  dhe_keygen(rand_sec);
  
  fprintf(stdout, "\n===== DHE RANDOM SECRET GENERATION TEST=====\n\n");

  mpz_get_str(sec_str, 16, rand_sec);

  fprintf(stdout, "Random Secret: %s\n", sec_str);
  if (strlen(sec_str) == DHE_KEY_LENGTH * HEX_CHAR_PER_BYTE) {
    fprintf(stdout, "Random Secret has correct length.\n\n");
  } else {
    fprintf(stdout, "Random Secret has incorrect length.\n\n");
    passes = 0;
  }

  if (passes) {
    fprintf(stdout, "\n=== DHE KEY GENERATION TEST: PASS ===\n\n");
  } else {
    fprintf(stdout, "\n=== DHE KEY GENERATION TEST: FAIL ===\n\n");
  }
}


int main() {
  rsa_keygen_test();
  dhe_keygen_test();
};
